unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls;

const
  AM_STOP = WM_USER + $96;
  AM_ARRAYMSG = WM_USER + $97;
  AM_ENTRYMSG = WM_USER + $98;
  AM_ERRORMSG = WM_USER + $95;

type
  TMainForm = class(TForm)
    Panel1: TPanel;
    DeviceList: TListBox;
    Panel2: TPanel;
    DetectButton: TButton;
    Panel3: TPanel;
    Panel5: TPanel;
    RepLabel0: TLabel;
    RepLabel1: TLabel;
    RepLabel2: TLabel;
    StatusList: TListBox;
    ReadClockTrack: TTrackBar;
    ReadClockLabel: TLabel;
    RepLabel3: TLabel;
    StartInButton: TButton;
    StopButton: TButton;
    RepLabel4: TLabel;
    StopCheck: TCheckBox;
    ClearNoRadio: TRadioButton;
    Clear1Radio: TRadioButton;
    Clear86Radio: TRadioButton;
    MegatransferEdit: TEdit;
    RepLabel5: TLabel;
    CountCheck: TCheckBox;
    Label1: TLabel;
    BlockEdit: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure DetectButtonClick(Sender: TObject);
    procedure DeviceListClick(Sender: TObject);
    procedure StartButtonClick(Sender: TObject);
    procedure StopButtonClick(Sender: TObject);
    procedure ReadClockTrackChange(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
  protected
    procedure AMArrayMsg(var Msg: TMessage); message AM_ARRAYMSG;
    procedure AMEntryMsg(var Msg: TMessage); message AM_ENTRYMSG;
    procedure AMErrorMsg(var Msg: TMessage); message AM_ERRORMSG;
    procedure AMStop(var Msg: TMessage); message AM_STOP;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  AIOUSB;

//const
  //BlockSize = LongInt(128 * 1024); //In transfers
  //BlockSize = LongInt(1024 * 1024); //In transfers
  //BlockSize = LongInt(8192); //In transfers
  //BlockSize = LongInt(2048); //In transfers
  //BlockSize = LongInt(8192*7*4); //In transfers

type
  TReal = Single;

  TStreamThread = class(TThread) //The main worker thread
  protected
    procedure Execute; override;
  public
    hMsgMutex: THandle;
    MsgVal: LongWord;
    bDoStop: Boolean;
    ClearFIFOMode: LongWord;
  end;

  TWriteThread = class(TThread) //Produces blank buffers for the stream thread to take data into
  public
    bSweepData: Boolean;
    hInSemaphore, hOutSemaphore,
    hBufMutex: THandle;
    Buf: array[Byte] of Pointer;
    BufIn, BufOut: Byte;
  protected
    procedure Execute; override;
  end;

  TReadThread = class(TThread) //Processes (and frees) buffers
  public
    bSweepData, bCheckData: Boolean;
    hInSemaphore, hOutSemaphore,
    hBufMutex: THandle;
    Buf: array[Byte] of Pointer;
    BufIn, BufOut: Byte;
  protected
    procedure Execute; override;
  end;

var
  DeviceIndex: LongWord;
  StreamThread: TStreamThread;
  WriteThread: TWriteThread;
  ReadThread:  TReadThread;
  TicksPerSec: Int64;
  TimeFactor: TReal;
  bFirst: Boolean;
  SmoothedTime: TReal;
  MegatransfersTotal: Integer;
  bCheckFirst: Boolean;
  CheckExpected: Word;
  CheckBlock: Integer;
  BlockSize: Integer = 8192;





function AIOUSB_ClearFIFO(DeviceIndex: LongWord; TimeMethod: LongWord): LongWord; cdecl; external 'AIOUSB.dll';

{ TStreamThread }

procedure TStreamThread.Execute;
var
  oTime, nTime: Int64;
  StopTime: Int64;
  dTime, Ret, L: LongWord;
  Block: Pointer;
  TransfersDone: Integer;

begin
  SetThreadPriority(GetCurrentThread, THREAD_PRIORITY_HIGHEST);

  TransfersDone := 0;
  QueryPerformanceCounter(StopTime);
  StopTime := StopTime + TicksPerSec * 3;
  QueryPerformanceCounter(nTime);
  repeat
    oTime := nTime;

    WaitForSingleObject(WriteThread.hOutSemaphore, INFINITE);
    WaitForSingleObject(WriteThread.hBufMutex, INFINITE);
      Block := WriteThread.Buf[WriteThread.BufOut];
      Inc(WriteThread.BufOut);
    ReleaseMutex(WriteThread.hBufMutex);
    ReleaseSemaphore(WriteThread.hInSemaphore, 1, nil);

    if ClearFIFOMode <> 0 then
      AIOUSB_ClearFIFO(DeviceIndex, ClearFIFOMode)
    ;

    Ret := DIO_StreamFrame(DeviceIndex, BlockSize, Block, L);

    QueryPerformanceCounter(nTime);

    oTime := nTime - oTime;
    if oTime > High(LongWord) then
      dTime := High(LongWord)
    else
      dTime := oTime
    ;

    L := L div 2;
    Dec(PLongWord(Block));
    PLongWord(Block)^ := L;
    Inc(PLongWord(Block));

    WaitForSingleObject(ReadThread.hBufMutex, INFINITE);
      ReadThread.Buf[ReadThread.BufIn] := Block;
      Inc(ReadThread.BufIn);
    ReleaseMutex(ReadThread.hBufMutex);
    ReleaseSemaphore(ReadThread.hInSemaphore, 1, nil);

    PostMessage(MainForm.Handle, AM_ARRAYMSG, dTime, Ret);
    {
    if bDoStop and ((nTime - StopTime) > 0) then begin
      Terminate;
      PostMessage(MainForm.Handle, AM_STOP, 0, 0);
    end;
    }
    if bDoStop then begin
      Inc(TransfersDone, L);
      if TransfersDone >= MegatransfersTotal * 1024*1024 then begin
        Terminate;
        PostMessage(MainForm.Handle, AM_STOP, 0, 0);
      end;
    end;
  until Terminated;
end;

{ TWriteThread }

procedure TWriteThread.Execute;
var
  Block: PLongWord;

begin
  repeat
    WaitForSingleObject(hInSemaphore, INFINITE);
    if Terminated then Break;

    GetMem(Block, BlockSize * 2 + 4);
    ZeroMemory(Block, BlockSize * 2 + 4);
    Inc(Block);

    WaitForSingleObject(hBufMutex, INFINITE);
      Buf[BufIn] := Block;
      Inc(BufIn);
    ReleaseMutex(hBufMutex);

    ReleaseSemaphore(hOutSemaphore, 1, nil);

  until Terminated;
end;

function IntToBin(I, MinBits: LongWord): String;
var
  DoneBits: LongWord;

begin
  Result := '';
  DoneBits := 0;
  repeat
    Result := Char(Byte('0') + (I and 1)) + Result;
    I := I shr 1;
    Inc(DoneBits);
  until (I = 0) and (DoneBits >= MinBits);
end;

{
function IntToBin(Value: Integer; Digits: Integer): string;
var i:integer;
begin
  result:='';
  for i:=((Digits*4)-1) downto 0 do begin
    result:=result+chr(((Value shr i) and 1)+$30);
  end;//for
end;
}

procedure fBinToHexView(BufList: TList; Stream: TStream);
var
  I, J, K, L, LastErrorJ: Integer;
  pBuf4: PLongWord;
  pBuf2: PWord;
  nData, xData: Word;
  SuBuf, Result, ExStr: String;
  Exceptions: LongWord;

begin
  Result := '';
  ExStr := '';
  Exceptions := 0;
  LastErrorJ := 0;
  J := 0;
  xData := 0;
  for I := 0 to BufList.Count - 1 do begin
    pBuf4 := BufList[I];
    Dec(pBuf4);
    L := pBuf4^;
    Inc(pBuf4);
    pBuf2 := PWord(pBuf4);
    for K := 0 to L - 1 do begin
      SuBuf := '';
      if (J mod 1024) = 0 then begin
        Stream.Write(Result[1], Length(Result));
        Result := '';
        PostMessage(MainForm.Handle, AM_ENTRYMSG, $10, J div BlockSize);
      end;
      if (K mod 512) = 0 then begin
        SuBuf := SuBuf + ExStr + #13#10;
        ExStr := '';
        if (J mod 16) <> 0 then SuBuf := SuBuf + StringOfChar(' ', 9 + (J mod 16) * 5);
      end;
      //if (J mod BlockSize) = 0 then SuBuf := SuBuf + #13#10;
      if (J mod 16) = 0 then begin
        SuBuf := SuBuf + ExStr;
        ExStr := '';
        SuBuf := SuBuf + #13#10 + IntToHex(J, 8) + ': ';
      end
      else
        SuBuf := SuBuf + ' '
      ;
      nData := pBuf2^;
      Inc(pBuf2);
      SuBuf := SuBuf + IntToHex(nData, 4);
      if nData <> xData then begin
        if ExStr = '' then ExStr := ' <----';
        ExStr := ExStr + ' [expected ' + IntToHex(xData, 4) + ', got ' + IntToHex(nData, 4) + ', been ' + IntToStr(J - LastErrorJ) + ']';
        Inc(Exceptions);
        LastErrorJ := J;
      end;
      xData := nData + 1;

      Result := Result + SuBuf;
      Inc(J);
    end;
    if L <> BlockSize then
      Result := Result + ExStr + #13#10'[End of ' + IntToHex(L, 1) + 'h buffer.]'#13#10#13#10
    else
      Result := Result + ExStr + #13#10#13#10#13#10
    ;
    ExStr := '';
    if (J mod 16) <> 0 then Result := Result + StringOfChar(' ', 9 + (J mod 16) * 5);
  end;
  Stream.Write(Result[1], Length(Result));
  Result := #13#10#13#10 + IntToStr(Exceptions) + ' errors detected.';
  Stream.Write(Result[1], Length(Result));
  Result := '';
  PostMessage(MainForm.Handle, AM_ENTRYMSG, $10, J div BlockSize);
end;

{ TReadThread }

procedure TReadThread.Execute;
var
  Block: PLongWord;
  I, L: Integer;
  OutFil: TFileStream;
  OutList: TList;
  pBuf: PWord;

begin
  OutList := nil; //For the compiler
  if bSweepData then begin
    OutList := TList.Create;
    OutList.Capacity := 64;
  end;

  repeat
    PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 0);
    WaitForSingleObject(hInSemaphore, INFINITE);
    PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 1);
    if Terminated then Break;
    PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 2);

    WaitForSingleObject(hBufMutex, INFINITE);
      PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 3);
      Block := Buf[BufOut];
      PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 4);
      Inc(BufOut);
      PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 5);
    ReleaseMutex(hBufMutex);
    PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 6);

    if bCheckData then begin
      Dec(Block);
      L := Block^;
      Inc(Block);
      pBuf := PWord(Block);
      if bCheckFirst then begin
        bCheckFirst := False;
        CheckExpected := pBuf^;
      end;
      for I := 0 to L - 1 do begin
        if pBuf^ <> CheckExpected then begin
          PostMessage(MainForm.Handle, AM_ERRORMSG, CheckBlock, I);
          CheckExpected := pBuf^ + 1;
        end
        else begin
          Inc(CheckExpected);
        end;
        Inc(pBuf);
      end;
      Inc(CheckBlock);
    end;

    if bSweepData then begin
      PostMessage(MainForm.Handle, AM_ENTRYMSG, $B, 0);
      OutList.Add(Block);
    end
    else begin
      PostMessage(MainForm.Handle, AM_ENTRYMSG, $C, 0);
      Dec(Block);
      FreeMem(Block, BlockSize * 2 + 4); //Memory management closes the loop, reusing freed buffers of the same size by preference.
    end;
    PostMessage(MainForm.Handle, AM_ENTRYMSG, $A, 7);

  until Terminated;

  PostMessage(MainForm.Handle, AM_ENTRYMSG, 0, 0);

  if bSweepData then begin
    PostMessage(MainForm.Handle, AM_ENTRYMSG, 4, 0);
    OutFil := TFileStream.Create('.\ThroughDat.txt', fmCreate);
    PostMessage(MainForm.Handle, AM_ENTRYMSG, 4, 1);
    fBinToHexView(OutList, OutFil);
    PostMessage(MainForm.Handle, AM_ENTRYMSG, 4, 2);
    OutFil.Free;
    PostMessage(MainForm.Handle, AM_ENTRYMSG, 4, 3);
    for I := 0 to OutList.Count - 1 do begin
      Block := OutList[I];
      Dec(Block);
      FreeMem(Block, BlockSize * 2 + 4);
      PostMessage(MainForm.Handle, AM_ENTRYMSG, 5, I);
    end;
    OutList.Free;
    PostMessage(MainForm.Handle, AM_ENTRYMSG, 4, 4);
  end;
end;





function ChunkTo(Len: Integer; S: String): String;
begin
  Result := '';
  while Length(S) > Len do begin
    Result := ' ' + Copy(S, Length(S) - Len + 1, Len) + Result;
    S := Copy(S, 1, Length(S) - Len);
  end;
  Result := S + Result;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  SetPriorityClass(GetCurrentProcess, HIGH_PRIORITY_CLASS);

  QueryPerformanceFrequency(TicksPerSec);
  TimeFactor := BlockSize * TicksPerSec;
  DetectButton.Click;
  ReadClockTrackChange(nil);
end;

procedure TMainForm.DetectButtonClick(Sender: TObject);
var
  DevMask: LongWord;
  I, Sel: Integer;
  NS, PID: LongWord;
  N: String;

begin
  DevMask := GetDevices;

  Sel := -1;

  DeviceList.Items.Clear;
  I := 0;
  while DevMask <> 0 do begin
    if (DevMask and 1) <> 0 then begin
      NS := 256;
      SetLength(N, NS);
      QueryDeviceInfo(I, @PID, @NS, PChar(N), nil, nil);
      SetLength(N, NS);
      Sel := DeviceList.Items.Add(IntToStr(I) + ': ' + N + '[' + IntToHex(PID, 4) + ']');
    end;

    DevMask := DevMask shr 1;
    Inc(I);
  end;
  DeviceList.ItemIndex := Sel;
  DeviceList.OnClick(nil);
end;

procedure TMainForm.DeviceListClick(Sender: TObject);
var
  I: Integer;
  Buf: String;

begin
  I := DeviceList.ItemIndex;
  if I < 0 then
    DeviceIndex := $FFFFFFFF
  else begin
    Buf := DeviceList.Items[I];
    I := Pos(':', Buf);
    Buf := Copy(Buf, 1, I - 1);
    DeviceIndex := StrToInt(Buf);
  end;
end;

procedure TMainForm.StartButtonClick(Sender: TObject);
var
  ReadClockHz, WriteClockHz: Double;
  OutMask, TriMask: Byte;
  DIOData: LongWord;

begin
  if DeviceIndex = $FFFFFFFF then begin
    Application.MessageBox('Please select a board(on the left) first.', 'No Selection', MB_ICONEXCLAMATION);
    Exit;
  end;

  BlockSize := StrToInt(BlockEdit.Text) * 1024;
  MegatransfersTotal := StrToInt(MegatransferEdit.Text);
  TimeFactor := BlockSize * TicksPerSec;

  StatusList.Items.Clear;
  StatusList.ItemIndex := StatusList.Items.Add('Starting...');
  DetectButton.Enabled := False;
  DeviceList.Enabled := False;
  StartInButton.Enabled := False;
  ReadClockTrack.Enabled := False;
  StopCheck.Enabled := False;

  RepLabel0.Caption := '';
  RepLabel1.Caption := '';
  RepLabel2.Caption := '';
  RepLabel3.Caption := '';
  RepLabel4.Caption := '';

  StatusList.ItemIndex := StatusList.Items.Add('Setting Clocks');
  ReadClockHz := ReadClockTrack.Position * 1000;
  WriteClockHz := ReadClockHz;
  StatusList.ItemIndex := StatusList.Items.Add(format('Clocks=%9.1f',[ReadClockHz]));
  DIO_StreamSetClocks(DeviceIndex, ReadClockHz, WriteClockHz);

  StatusList.ItemIndex := StatusList.Items.Add('StreamOpen');
  DIO_StreamOpen(DeviceIndex, True);
  OutMask := $08;
  DIOData := $FFFFFFFF;
  TriMask := $01;
  DIO_ConfigureEx(DeviceIndex, @OutMask, @DIOData, @TriMask);
  bFirst := True;
  StatusList.ItemIndex := StatusList.Items.Add('Creating Write Thread');
  WriteThread := TWriteThread.Create(True);
  with WriteThread do begin
    FreeOnTerminate := False;
    hInSemaphore := CreateSemaphore(nil, 30, 32, nil);
    hOutSemaphore := CreateSemaphore(nil, 0, 32, nil);
    hBufMutex := CreateMutex(nil, False, nil);
    BufIn := 0;
    BufOut := 0;
    Resume;
  end;
  bCheckFirst := True;
  CheckBlock := 0;
  StatusList.ItemIndex := StatusList.Items.Add('Creating Read Thread');
  ReadThread := TReadThread.Create(True);
  with ReadThread do begin
    FreeOnTerminate := False;
    bSweepData := StopCheck.Checked;
    bCheckData := CountCheck.Checked;
    hInSemaphore := CreateSemaphore(nil, 0, 512, nil);
    hOutSemaphore := 0;
    hBufMutex := CreateMutex(nil, False, nil);
    BufIn := 0;
    BufOut := 0;
    Resume;
  end;
  StatusList.ItemIndex := StatusList.Items.Add('Creating "Stream" Thread');
  StreamThread := TStreamThread.Create(True);
  with StreamThread do begin
    FreeOnTerminate := False;
    hMsgMutex := CreateMutex(nil, False, nil);
    bDoStop := StopCheck.Checked;
    if ClearNoRadio.Checked then
      ClearFIFOMode := 0
    else if Clear1Radio.Checked then
      ClearFIFOMode := 1
    else
      ClearFIFOMode := 86
    ;
  end;
  StatusList.ItemIndex := StatusList.Items.Add('Waiting for Write Thread to allocate buffers');
  while WaitForSingleObject(WriteThread.hInSemaphore, 0) = WAIT_OBJECT_0 do begin //Wait until the write thread has all buffers ready
    ReleaseSemaphore(WriteThread.hInSemaphore, 1, nil);
    Sleep(1);
  end;
  StatusList.ItemIndex := StatusList.Items.Add('Done Waiting for Write Thread to allocate buffers');
  StatusList.ItemIndex := StatusList.Items.Add('Resuming StreamThread');

  StreamThread.Resume;
  StatusList.ItemIndex := StatusList.Items.Add('Done Resuming StreamThread');

  StopButton.Enabled := True;
  StatusList.ItemIndex := StatusList.Items.Add('Started.');
end;


procedure TMainForm.StopButtonClick(Sender: TObject);
var
  ReadClockHz, WriteClockHz: Double;
  I: Integer;

begin
  if not StopButton.Enabled then Exit;
  StatusList.ItemIndex := StatusList.Items.Add('Stopping...');
  StopButton.Enabled := False;

  StatusList.ItemIndex := StatusList.Items.Add('Preflushing read thread...');
  StatusList.ItemIndex := StatusList.Items.Add('Terminating read thread...');
  ReadThread.Terminate;
  StatusList.ItemIndex := StatusList.Items.Add('Strobing read thread...');
  for I := 0 to 41 do
    ReleaseSemaphore( ReadThread.hInSemaphore, 1, nil)
  ;
  StatusList.ItemIndex := StatusList.Items.Add('Waiting for read thread to flush...');
  //ReadThread.WaitFor;
  while WaitForSingleObject(ReadThread.Handle, 1) = WAIT_TIMEOUT do Application.ProcessMessages;
  StatusList.ItemIndex := StatusList.Items.Add('Preflushed read thread.');

  StatusList.ItemIndex := StatusList.Items.Add('Stopping stream thread...');
  StatusList.ItemIndex := StatusList.Items.Add('Terminating stream thread...');
  StreamThread.Terminate;
  StatusList.ItemIndex := StatusList.Items.Add('Waiting for stream thread to close...');
  StreamThread.WaitFor;
  StatusList.ItemIndex := StatusList.Items.Add('Freeing stream thread...');
  StreamThread.Free;
  StatusList.ItemIndex := StatusList.Items.Add('Stopping clocks...');
  ReadClockHz := 0;
  WriteClockHz := 0;
  DIO_StreamSetClocks(DeviceIndex, ReadClockHz, WriteClockHz);
  StatusList.ItemIndex := StatusList.Items.Add('Closing stream process...');
  DIO_StreamClose(DeviceIndex);

  StatusList.ItemIndex := StatusList.Items.Add('Stopping write thread...');
  WriteThread.Terminate;
  ReleaseSemaphore(WriteThread.hInSemaphore, 1, nil);
  WriteThread.WaitFor;
  CloseHandle(WriteThread.hInSemaphore);
  CloseHandle(WriteThread.hOutSemaphore);
  CloseHandle(WriteThread.hBufMutex);
  WriteThread.Free;
  StatusList.ItemIndex := StatusList.Items.Add('Stopping read thread...');
  ReadThread.Terminate;
  ReleaseSemaphore( ReadThread.hInSemaphore, 1, nil);
  ReadThread.WaitFor;
  CloseHandle(ReadThread.hInSemaphore);
  CloseHandle(ReadThread.hBufMutex);
  ReadThread.Free;

  DetectButton.Enabled := True;
  DeviceList.Enabled := True;
  StartInButton.Enabled := True;
  ReadClockTrack.Enabled := True;
  StopCheck.Enabled := True;
  StatusList.ItemIndex := StatusList.Items.Add('Stopped.');
end;

procedure TMainForm.AMStop(var Msg: TMessage);
begin
  StopButton.Click;
end;

procedure TMainForm.AMArrayMsg(var Msg: TMessage);
var
  Buf: LongWord;

begin
  Buf := Msg.LParam;
  if Buf <> 0 then Caption := IntToStr(Buf);

  Buf := Msg.WParam;
  RepLabel0.Caption := 'Throughput: ' + ChunkTo(3, IntToStr(Round(TimeFactor / Buf))) + ' T/s';
  if Buf = $FFFFFFFF then
    RepLabel2.Caption := 'Interval too long'
  else begin
    RepLabel2.Caption := 'Interval: ' + Format('%.8g', [ Buf / TicksPerSec ]) + 's';
    if bFirst then begin
      bFirst := False;
      SmoothedTime := Buf;
    end
    else begin
      SmoothedTime := SmoothedTime * 0.99 + Buf * 0.01;
    end;
    RepLabel1.Caption := '(Smoothed): ' + ChunkTo(3, IntToStr(Round(TimeFactor / SmoothedTime))) + ' T/s';
    RepLabel3.Caption := '(Smoothed): ' + Format('%.8g', [ SmoothedTime / TicksPerSec ]) + 's';
  end;
end;

procedure TMainForm.ReadClockTrackChange(Sender: TObject);
var
  I: Integer;

begin
  I := ReadClockTrack.Position;
  if I = 0 then
    ReadClockLabel.Caption := '&Input Clock: External'
  else
    ReadClockLabel.Caption := '&Input Clock: ' + IntToStr(I) + ' KHz'
  ;
end;

procedure TMainForm.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  if StopButton.Enabled then begin
    case Application.MessageBox('Stop before closing?', 'Stop?', MB_ICONQUESTION or MB_YESNOCANCEL) of
      ID_YES: begin
        StopButton.Click;
        CanClose := True;
      end;
      ID_NO: begin
        CanClose := True;
      end;
      else begin
        CanClose := False;
      end;
    end;
  end;
end;

procedure TMainForm.AMEntryMsg(var Msg: TMessage);
begin
  RepLabel4.Caption := IntToHex(Msg.WParam, 1) + ':' + IntToHex(Msg.LParam, 1);
end;

procedure TMainForm.AMErrorMsg(var Msg: TMessage);
begin
  StatusList.Items.Add('Error at ' + IntToHex(Msg.WParam, 1) + ':' + IntToHex(Msg.LParam div 512, 1) + ':' + IntToHex(Msg.LParam mod 512, 1));
end;

end.
